from serpent.window_controller import WindowController

import subprocess
import shlex

import re


class LinuxWindowController(WindowController):

    def __init__(self):
        pass

    def locate_window(self, name):
        return subprocess.check_output(shlex.split(f"xdotool search --name \"^{name}$\"")).decode("utf-8").strip()

    def move_window(self, window_id, x, y):
        subprocess.call(shlex.split(f"xdotool windowmove {window_id} {x} {y}"))

    def resize_window(self, window_id, width, height):
        subprocess.call(shlex.split(f"xdotool windowsize {window_id} {width} {height}"))

    def focus_window(self, window_id):
        subprocess.call(shlex.split(f"xdotool windowactivate {window_id}"))

    def is_window_focused(self, window_id):
        focused_window_id = subprocess.check_output(shlex.split("xdotool getwindowfocus")).decode("utf-8").strip()
        return focused_window_id == window_id

    def get_focused_window_name(self):
        focused_window_id = subprocess.check_output(shlex.split("xdotool getwindowfocus")).decode("utf-8").strip()
        return subprocess.check_output(shlex.split(f"xdotool getwindowname {focused_window_id}")).decode("utf-8").strip()

    def get_window_geometry(self, window_id):
        geometry = dict()

        window_geometry = subprocess.check_output(shlex.split(f"xdotool getwindowgeometry {window_id}")).decode("utf-8").strip()
        size = re.match(r"\s+Geometry: ([0-9]+x[0-9]+)", window_geometry.split("\n")[2]).group(1).split("x")

        geometry["width"] = int(size[0])
        geometry["height"] = int(size[1])

        window_information = subprocess.check_output(shlex.split(f"xwininfo -id {window_id}")).decode("utf-8").strip()
        geometry["x_offset"] = int(re.match(r"\s+Absolute upper-left X:\s+([0-9]+)", window_information.split("\n")[2]).group(1))
        geometry["y_offset"] = int(re.match(r"\s+Absolute upper-left Y:\s+([0-9]+)", window_information.split("\n")[3]).group(1))

        return geometry



from serpent.machine_learning.reinforcement_learning.dqn import DQN

import numpy as np

from keras.optimizers import Adam

import random


class DDQN(DQN):

    def __init__(
        self,
        input_shape=None,
        input_mapping=None,
        replay_memory_size=10000,
        batch_size=32,
        action_space=None,
        max_steps=1000000,
        observe_steps=None,
        initial_epsilon=1.0,
        final_epsilon=0.1,
        gamma=0.99,
        model_file_path=None,
        model_learning_rate=1e-4,
        override_epsilon=False
    ):
        super().__init__(
            input_shape=input_shape,
            input_mapping=input_mapping,
            replay_memory_size=replay_memory_size,
            batch_size=batch_size,
            action_space=action_space,
            max_steps=max_steps,
            observe_steps=observe_steps,
            initial_epsilon=initial_epsilon,
            final_epsilon=final_epsilon,
            gamma=gamma,
            model_file_path=model_file_path,
            model_learning_rate=model_learning_rate,
            override_epsilon=override_epsilon
        )

        self.type = "DDQN"
        self.model_online = self._initialize_model()

    def calculate_target_error(self, observation):
        previous_target = self.model_online.predict(observation[0])[0][observation[1]]

        if observation[4]:
            target = observation[2]
        else:
            best_action = np.argmax(self.model_online.predict(observation[3]))
            q = self.model.predict(observation[3])[0][best_action]

            target = observation[2] + self.gamma * q

        return np.abs(target - previous_target)

    def train_on_mini_batch(self):
        mini_batch = self.generate_mini_batch()

        flashback_indices = random.sample(range(self.batch_size), 6)

        for i in range(0, len(mini_batch)):
            if i in flashback_indices:
                flashback_image = np.squeeze(mini_batch[i][1][3][:, :, :, 1])

                self.visual_debugger.store_image_data(
                    np.array(flashback_image * 255, dtype="uint8"),
                    flashback_image.shape,
                    f"flashback_{flashback_indices.index(i) + 1}"
                )

                del flashback_image

            previous_frame_stack = mini_batch[i][1][0]
            action_index = mini_batch[i][1][1]
            reward = mini_batch[i][1][2]
            frame_stack = mini_batch[i][1][3]
            terminal = mini_batch[i][1][4]

            target = self.model_online.predict(previous_frame_stack)
            previous_target = target[0][action_index]

            if terminal:
                target[0][action_index] = reward
            else:
                best_action = np.argmax(self.model_online.predict(frame_stack))
                q = self.model.predict(frame_stack)[0][best_action]

                target[0][action_index] = reward + self.gamma * q

            error = np.abs(target[0][action_index] - previous_target)
            self.replay_memory.update(mini_batch[i][0], error)

            self.model_online.fit(previous_frame_stack, target, epochs=1, verbose=0)

        self.model_loss = 0

    def pick_action(self, action_type=None):
        if action_type is None:
            self.compute_action_type()
        else:
            self.current_action_type = action_type

        qs = self.model_online.predict(self.frame_stack)

        if self.current_action_type == "RANDOM":
            self.current_action_index = random.randrange(self.action_count)
            self.maximum_future_rewards = None
        elif self.current_action_type == "PREDICTED":
            self.current_action_index = np.argmax(qs)
            self.maximum_future_rewards = qs

    def save_model_weights(self, file_path_prefix="datasets/model_", is_checkpoint=False):
        epsilon = self.epsilon_greedy_q_policy.epsilon

        if is_checkpoint:
            file_path = f"{file_path_prefix}_dqn_{self.current_step}_{epsilon}_.h5"
        else:
            file_path = f"{file_path_prefix}_dqn_{epsilon}_.h5"

        self.model_online.save_weights(file_path, overwrite=True)

    def load_model_weights(self, file_path, override_epsilon):
        self.model_online.load_weights(file_path)
        self.model_online.compile(loss="logcosh", optimizer=Adam(lr=self.model_learning_rate, clipvalue=10))

        self.update_target_model()

        *args, steps, epsilon, extension = file_path.split("_")
        self.current_step = int(steps)

        if override_epsilon:
            self.previous_epsilon = float(epsilon)
            self.epsilon_greedy_q_policy.epsilon = float(epsilon)

    def update_target_model(self):
        print("Updating target model...")
        self.model.set_weights(self.model_online.get_weights())



from serpent.game import Game

from .api.api import MyGameAPI

from serpent.utilities import Singleton

from serpent.game_launchers.web_browser_game_launcher import WebBrowser


class SerpentGame(Game, metaclass=Singleton):

    def __init__(self, **kwargs):
        kwargs["platform"] = "PLATFORM"

        kwargs["window_name"] = "WINDOW_NAME"

        kwargs["app_id"] = "APP_ID"
        kwargs["app_args"] = None
        kwargs["executable_path"] = "EXECUTABLE_PATH"
        kwargs["url"] = "URL"
        kwargs["browser"] = WebBrowser.DEFAULT

        super().__init__(**kwargs)

        self.api_class = MyGameAPI
        self.api_instance = None

    @property
    def screen_regions(self):
        regions = {
            "SAMPLE_REGION": (0, 0, 0, 0)
        }

        return regions

    @property
    def ocr_presets(self):
        presets = {
            "SAMPLE_PRESET": {
                "extract": {
                    "gradient_size": 1,
                    "closing_size": 1
                },
                "perform": {
                    "scale": 10,
                    "order": 1,
                    "horizontal_closing": 1,
                    "vertical_closing": 1
                }
            }
        }

        return presets



import offshoot

from serpent.config import config

import time
import uuid
import pickle

import os
import os.path

import serpent.cv
import serpent.ocr
import serpent.utilities

from serpent.frame_grabber import FrameGrabber
from serpent.game_frame_buffer import GameFrameBuffer
from serpent.sprite_identifier import SpriteIdentifier
from serpent.visual_debugger.visual_debugger import VisualDebugger

import skimage.io
import skimage.transform

from redis import StrictRedis

from datetime import datetime


class GameAgentError(BaseException):
    pass


class GameAgent(offshoot.Pluggable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.game = kwargs["game"]
        self.game.api

        self.config = config.get(f"{self.__class__.__name__}Plugin") or dict()

        self.redis_client = StrictRedis(**config["redis"])

        self.input_controller = kwargs["input_controller"]
        self.machine_learning_models = dict()

        self.frame_handlers = dict(
            NOOP=self.handle_noop,
            COLLECT_FRAMES=self.handle_collect_frames,
            COLLECT_FRAME_REGIONS=self.handle_collect_frame_regions,
            COLLECT_FRAMES_FOR_CONTEXT=self.handle_collect_frames_for_context
        )

        self.frame_handler_setups = dict(
            COLLECT_FRAMES=self.setup_collect_frames,
            COLLECT_FRAME_REGIONS=self.setup_collect_frame_regions,
            COLLECT_FRAMES_FOR_CONTEXT=self.setup_collect_frames_for_context
        )

        self.frame_handler_pause_callbacks = dict(
            COLLECT_FRAMES=self.on_collect_frames_pause,
            COLLECT_FRAME_REGIONS=self.on_collect_frame_regions_pause,
            COLLECT_FRAMES_FOR_CONTEXT=self.on_collect_frames_for_context_pause
        )

        self.frame_handler_setup_performed = False

        self.visual_debugger = VisualDebugger()

        self.game_frame_buffer = GameFrameBuffer(size=self.config.get("game_frame_buffer_size", 5))
        self.game_context = None

        self.sprite_identifier = SpriteIdentifier()
        self._register_sprites()

        self.flag = None

        self.uuid = str(uuid.uuid4())
        self.started_at = datetime.now()

    @offshoot.forbidden
    def on_game_frame(self, game_frame, frame_handler=None, **kwargs):
        if not self.frame_handler_setup_performed:
            self._setup_frame_handler(frame_handler=frame_handler, **kwargs)

        frame_handler = self.frame_handlers.get(frame_handler or self.config.get("frame_handler", "NOOP"))

        frame_handler(game_frame, **kwargs)

        self.game_frame_buffer.add_game_frame(game_frame)

    @offshoot.forbidden
    def on_pause(self, frame_handler=None, **kwargs):
        on_frame_handler_pause = self.frame_handler_pause_callbacks.get(frame_handler or self.config.get("frame_handler"))

        if on_frame_handler_pause is not None:
            on_frame_handler_pause(**kwargs)

    @offshoot.forbidden
    def load_machine_learning_model(self, file_path):
        with open(file_path, "rb") as f:
            serialized_classifier = f.read()

        return pickle.loads(serialized_classifier)

    @offshoot.forbidden
    def update_game_frame(self, game_frame):
        game_frame_buffer = FrameGrabber.get_frames([0], frame_shape=game_frame.frame.shape)
        return game_frame_buffer.frames[0]

    def handle_noop(self, game_frame, **kwargs):
        time.sleep(1)

    def setup_collect_frames(self, **kwargs):
        self.game_frames = list()
        self.collected_frame_count = 0

    def setup_collect_frame_regions(self, **kwargs):
        self.game_frames = list()
        self.collected_frame_count = 0

    def setup_collect_frames_for_context(self, **kwargs):
        context = kwargs.get("context") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["context"]

        if not os.path.isdir(f"datasets/collect_frames_for_context/{context}"):
            os.mkdir(f"datasets/collect_frames_for_context/{context}")

        self.game_frames = list()
        self.collected_frame_count = 0

    def handle_collect_frames(self, game_frame, **kwargs):
        self.game_frames.append(game_frame)

        self.collected_frame_count += 1

        serpent.utilities.clear_terminal()
        print(f"Collected Frame #{self.collected_frame_count}")

        time.sleep(kwargs.get("interval") or self.config.get("collect_frames_interval") or 1)

    def handle_collect_frame_regions(self, game_frame, **kwargs):
        region = kwargs.get("region")

        self.game_frames.append(game_frame)

        self.collected_frame_count += 1

        serpent.utilities.clear_terminal()
        print(f"Collected Frame #{self.collected_frame_count} for Region: {region}")

        time.sleep(kwargs.get("interval") or self.config.get("collect_frames_interval") or 1)

    def handle_collect_frames_for_context(self, game_frame, **kwargs):
        context = kwargs.get("context") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["context"]
        interval = kwargs.get("interval") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["interval"]

        self.game_frames.append(game_frame)

        self.collected_frame_count += 1

        serpent.utilities.clear_terminal()
        print(f"Collected Frame #{self.collected_frame_count} for Context: {context}")

        time.sleep(interval)

    def on_collect_frames_pause(self, **kwargs):
        for i, game_frame in enumerate(self.game_frames):
            print(f"Saving image {i + 1}/{len(self.game_frames)} to disk...")
            skimage.io.imsave(f"datasets/collect_frames/frame_{str(uuid.uuid4())}.png", game_frame.frame)

        self.game_frames = list()

    def on_collect_frame_regions_pause(self, **kwargs):
        region = kwargs["region"]
        region_path = f"datasets/collect_frames/{region}"

        for i, game_frame in enumerate(self.game_frames):
            frame_region = serpent.cv.extract_region_from_image(game_frame.frame, self.game.screen_regions.get(region))

            if not os.path.isdir(region_path):
                os.mkdir(region_path)

            print(f"Saving image {i + 1}/{len(self.game_frames)} to disk...")
            skimage.io.imsave(f"{region_path}/region_{str(uuid.uuid4())}.png", frame_region)

        self.game_frames = list()

    def on_collect_frames_for_context_pause(self, **kwargs):
        context = kwargs.get("context") or config["frame_handlers"]["COLLECT_FRAMES_FOR_CONTEXT"]["context"]

        for i, game_frame in enumerate(self.game_frames):
            resized_frame = skimage.transform.resize(
                game_frame.frame,
                (game_frame.frame.shape[0] // 2, game_frame.frame.shape[1] // 2)
            )

            file_name = f"datasets/collect_frames_for_context/{context}/frame_{str(uuid.uuid4())}.png"

            print(f"Saving image {i + 1}/{len(self.game_frames)} to disk...")
            skimage.io.imsave(file_name, resized_frame)

        self.game_frames = list()

    def _setup_frame_handler(self, frame_handler=None, **kwargs):
        frame_handler = frame_handler or self.config.get("frame_handler", "NOOP")

        if frame_handler in self.frame_handler_setups:
            self.frame_handler_setups[frame_handler](**kwargs)

        self.frame_handler_setup_performed = True

    def _register_sprites(self):
        for sprite_name, sprite in self.game.sprites.items():
            self.sprite_identifier.register(sprite)



import asyncio

from autobahn.asyncio.wamp import ApplicationSession, ApplicationRunner
from autobahn.wamp.types import RegisterOptions, SubscribeOptions
from autobahn.wamp import auth

from serpent.config import config

import aioredis
import json


class AnalyticsComponent:
    @classmethod
    def run(cls):
        print(f"Starting {cls.__name__}...")

        url = "ws://%s:%s" % (config["analytics"]["host"], config["analytics"]["port"])

        runner = ApplicationRunner(url=url, realm=config["analytics"]["realm"])
        runner.run(AnalyticsWAMPComponent)


class AnalyticsWAMPComponent(ApplicationSession):
    def __init__(self, c=None):
        super().__init__(c)

    def onConnect(self):
        self.join(config["analytics"]["realm"], ["wampcra"], config["analytics"]["auth"]["username"])

    def onDisconnect(self):
        print("Disconnected from Crossbar!")

    def onChallenge(self, challenge):
        secret = config["analytics"]["auth"]["password"]
        signature = auth.compute_wcs(secret.encode('utf8'), challenge.extra['challenge'].encode('utf8'))

        return signature.decode('ascii')

    async def onJoin(self, details):
        self.redis_client = await self._initialize_redis_client()

        while True:
            redis_key, event = await self.redis_client.brpop("SERPENT:AISAAC_MAZE:EVENTS")
            event = json.loads(event.decode("utf-8"))

            topic = event.pop("project_key")
            self.publish(topic, event)

    async def _initialize_redis_client(self):
        return await aioredis.create_redis(
            (config["redis"]["host"], config["redis"]["port"]),
            loop=asyncio.get_event_loop()
        )



import numpy as np

import random
import uuid


class SpriteError(BaseException):
    pass


class Sprite:

    def __init__(self, name, image_data=None, signature_colors=None, constellation_of_pixels=None):
        if not isinstance(image_data, np.ndarray):
            raise SpriteError("'image_data' needs to be a 4D instance of ndarray...")

        if not len(image_data.shape) == 4:
            raise SpriteError("'image_data' ndarray needs to be 4D...")

        self.name = name
        self.image_data = image_data
        self.image_shape = image_data.shape[:2]

        self.signature_colors = signature_colors or self._generate_signature_colors()

        self.constellation_of_pixels = constellation_of_pixels or self._generate_constellation_of_pixels()

    def append_image_data(self, image_data, signature_colors=None, constellation_of_pixels=None):
        images = list()

        for i in range(self.image_data.shape[3]):
            images.append((self.image_data[:, :, :, i])[:, :, :, np.newaxis])

        images.append(image_data)

        self.image_data = np.squeeze(np.stack(images, axis=3))

        if signature_colors is not None:
            self.signature_colors.append(signature_colors)
        else:
            self.signature_colors = self._generate_signature_colors()

        if constellation_of_pixels is not None:
            self.constellation_of_pixels.append(constellation_of_pixels)
        else:
            self.constellation_of_pixels = self._generate_constellation_of_pixels()

    def generate_constellation_of_pixels_images(self):
        constellation_of_pixel_images = list()

        for i in range(self.image_data.shape[3]):
            constellation_of_pixel_image = np.zeros(self.image_data[..., :3, 0].shape, dtype="uint8")

            for yx, rgb in self.constellation_of_pixels[0].items():
                constellation_of_pixel_image[yx[0], yx[1], :] = rgb

            constellation_of_pixel_images.append(constellation_of_pixel_image)

        return constellation_of_pixel_images

    def _generate_seed(self):
        return str(uuid.uuid4())

    def _generate_signature_colors(self, quantity=8):
        signature_colors = list()
        height, width, pixels, animation_states = self.image_data.shape

        for i in range(animation_states):
            values, counts = np.unique(self.image_data[..., i].reshape(width * height, pixels), axis=0, return_counts=True)

            if len(values[0]) == 3:
                maximum_indices = np.argsort(counts)[::-1][:quantity]
            elif len(values[0]) == 4:
                maximum_indices = list()

                for index in np.argsort(counts)[::-1]:
                    value = values[index]

                    if value[3] > 0:
                        maximum_indices.append(index)

                        if len(maximum_indices) == quantity:
                            break

            colors = [tuple(map(int, values[index][:3])) for index in maximum_indices]
            signature_colors.append(set(colors))

        return signature_colors

    def _generate_constellation_of_pixels(self, quantity=8):
        constellation_of_pixels = list()
        height, width, pixels, animation_states = self.image_data.shape

        for i in range(animation_states):
            constellation_of_pixels.append(dict())

            for ii in range(quantity):
                signature_color = random.choice(list(self.signature_colors[i]))
                signature_color_locations = Sprite.locate_color(signature_color, np.squeeze(self.image_data[:, :, :3, i]))

                y, x = random.choice(signature_color_locations)
                constellation_of_pixels[i][(y, x)] = signature_color

        return constellation_of_pixels

    @classmethod
    def locate_color(cls, color, image):
        # TODO: Optimize for ms gain

        if image.shape[2] == 3:
            color_indices = np.where(np.all(image[:, :, :3] == color, axis=-1))
        elif image.shape[2] == 4:
            color_indices = np.where(np.all(image[:, :, :3] == (list(color) + [255]), axis=-1))

        return list(zip(*color_indices)) if len(color_indices[0]) else list()



import numpy as np


def generate_rays(player_to_center_angle, mode="UNIFORM", quantity=12, starting_angle=30):
    rays = {"Ray Player": player_to_center_angle}

    clockwise = True
    current_angle = starting_angle
    current_angle_count = 0

    if mode == "UNIFORM":
        while len(rays) < quantity:
            ray_label = f"Ray Player {'-' if clockwise else '+'} {str(current_angle)}"

            if clockwise:
                rays[ray_label] = (player_to_center_angle - current_angle + 179) % 360 - 179
            else:
                rays[ray_label] = (player_to_center_angle + current_angle + 179) % 360 - 179

            current_angle_count += 1
            clockwise = not clockwise

            if current_angle_count == 2:
                current_angle_count = 0
                current_angle += starting_angle

    return rays


def calculate_minimum_collision_distances(rays, thresholded_frame, angle_mapping_array, distance_mapping_array):
    ray_collision_distances = dict()

    for label, angle in rays.items():
        ray_collision_mask = (
            (angle_mapping_array == angle) &
            (thresholded_frame == 1)
        )

        collision_distances = distance_mapping_array[ray_collision_mask == True]

        ray_collision_distances[label] = np.min(collision_distances) if collision_distances.size else 9999

    return ray_collision_distances



from redis import StrictRedis
from datetime import datetime

from serpent.config import config

import json


class AnalyticsClientError(BaseException):
    pass


class AnalyticsClient:

    def __init__(self, project_key=None):
        if project_key is None:
            raise AnalyticsClientError("'project_key' kwarg is expected...")

        self.project_key = project_key
        self.redis_client = StrictRedis(**config["redis"])

    @property
    def redis_key(self):
        return f"SERPENT:{self.project_key}:EVENTS"

    def track(self, event_key=None, data=None):
        event = {
            "project_key": self.project_key,
            "event_key": event_key,
            "data": data,
            "timestamp": datetime.utcnow().isoformat()
        }

        self.redis_client.lpush(self.redis_key, json.dumps(event))


from serpent.game_launchers.steam_game_launcher import SteamGameLauncher
from serpent.game_launchers.executable_game_launcher import ExecutableGameLauncher
from serpent.game_launchers.web_browser_game_launcher import WebBrowserGameLauncher


from serpent.game_api import GameAPI


class MyGameAPI(GameAPI):

    def __init__(self, game=None):
        super().__init__(game=game)

    def my_api_function(self):
        pass

    class MyAPINamespace:

        @classmethod
        def my_namespaced_api_function(cls):
            api = MyGameAPI.instance


from serpent.input_controller import InputController, MouseButton, KeyboardKey

from serpent.sprite_locator import SpriteLocator

import serpent.ocr

import pyautogui

import time


keyboard_key_mapping = {
    KeyboardKey.KEY_ESCAPE.name: "esc",
    KeyboardKey.KEY_F1.name: "f1",
    KeyboardKey.KEY_F2.name: "f2",
    KeyboardKey.KEY_F3.name: "f3",
    KeyboardKey.KEY_F4.name: "f4",
    KeyboardKey.KEY_F5.name: "f5",
    KeyboardKey.KEY_F6.name: "f6",
    KeyboardKey.KEY_F7.name: "f7",
    KeyboardKey.KEY_F8.name: "f8",
    KeyboardKey.KEY_F9.name: "f9",
    KeyboardKey.KEY_F10.name: "f10",
    KeyboardKey.KEY_F11.name: "f11",
    KeyboardKey.KEY_F12.name: "f12",
    KeyboardKey.KEY_PRINT_SCREEN.name: "printscreen",
    KeyboardKey.KEY_SCROLL_LOCK.name: "scrolllock",
    KeyboardKey.KEY_PAUSE.name: "pause",
    KeyboardKey.KEY_GRAVE.name: "`",
    KeyboardKey.KEY_BACKTICK.name: "`",
    KeyboardKey.KEY_1.name: "1",
    KeyboardKey.KEY_2.name: "2",
    KeyboardKey.KEY_3.name: "3",
    KeyboardKey.KEY_4.name: "4",
    KeyboardKey.KEY_5.name: "5",
    KeyboardKey.KEY_6.name: "6",
    KeyboardKey.KEY_7.name: "7",
    KeyboardKey.KEY_8.name: "8",
    KeyboardKey.KEY_9.name: "9",
    KeyboardKey.KEY_0.name: "0",
    KeyboardKey.KEY_MINUS.name: "-",
    KeyboardKey.KEY_DASH.name: "-",
    KeyboardKey.KEY_EQUALS.name: "=",
    KeyboardKey.KEY_BACKSPACE.name: "backspace",
    KeyboardKey.KEY_INSERT.name: "insert",
    KeyboardKey.KEY_HOME.name: "home",
    KeyboardKey.KEY_PAGE_UP.name: "pageup",
    KeyboardKey.KEY_NUMLOCK.name: "numlock",
    KeyboardKey.KEY_NUMPAD_DIVIDE.name: "divide",
    KeyboardKey.KEY_NUMPAD_SLASH.name: "divide",
    KeyboardKey.KEY_NUMPAD_MULTIPLY.name: "multiply",
    KeyboardKey.KEY_NUMPAD_STAR.name: "multiply",
    KeyboardKey.KEY_NUMPAD_SUBTRACT.name: "subtract",
    KeyboardKey.KEY_NUMPAD_DASH.name: "subtract",
    KeyboardKey.KEY_TAB.name: "tab",
    KeyboardKey.KEY_Q.name: "q",
    KeyboardKey.KEY_W.name: "w",
    KeyboardKey.KEY_E.name: "e",
    KeyboardKey.KEY_R.name: "r",
    KeyboardKey.KEY_T.name: "t",
    KeyboardKey.KEY_Y.name: "y",
    KeyboardKey.KEY_U.name: "u",
    KeyboardKey.KEY_I.name: "i",
    KeyboardKey.KEY_O.name: "o",
    KeyboardKey.KEY_P.name: "p",
    KeyboardKey.KEY_LEFT_BRACKET.name: "[",
    KeyboardKey.KEY_RIGHT_BRACKET.name: "]",
    KeyboardKey.KEY_BACKSLASH.name: "\\",
    KeyboardKey.KEY_DELETE.name: "delete",
    KeyboardKey.KEY_END.name: "end",
    KeyboardKey.KEY_PAGE_DOWN.name: "pagedown",
    KeyboardKey.KEY_NUMPAD_7.name: "num7",
    KeyboardKey.KEY_NUMPAD_8.name: "num8",
    KeyboardKey.KEY_NUMPAD_9.name: "num9",
    KeyboardKey.KEY_NUMPAD_ADD.name: "add",
    KeyboardKey.KEY_NUMPAD_PLUS.name: "add",
    KeyboardKey.KEY_CAPSLOCK.name: "capslock",
    KeyboardKey.KEY_A.name: "a",
    KeyboardKey.KEY_S.name: "s",
    KeyboardKey.KEY_D.name: "d",
    KeyboardKey.KEY_F.name: "f",
    KeyboardKey.KEY_G.name: "g",
    KeyboardKey.KEY_H.name: "h",
    KeyboardKey.KEY_J.name: "j",
    KeyboardKey.KEY_K.name: "k",
    KeyboardKey.KEY_L.name: "l",
    KeyboardKey.KEY_SEMICOLON.name: ';',
    KeyboardKey.KEY_APOSTROPHE.name: "'",
    KeyboardKey.KEY_RETURN.name: "enter",
    KeyboardKey.KEY_ENTER.name: "enter",
    KeyboardKey.KEY_NUMPAD_4.name: "num4",
    KeyboardKey.KEY_NUMPAD_5.name: "num5",
    KeyboardKey.KEY_NUMPAD_6.name: "num6",
    KeyboardKey.KEY_LEFT_SHIFT.name: "shiftleft",
    KeyboardKey.KEY_Z.name: "z",
    KeyboardKey.KEY_X.name: "x",
    KeyboardKey.KEY_C.name: "c",
    KeyboardKey.KEY_V.name: "v",
    KeyboardKey.KEY_B.name: "b",
    KeyboardKey.KEY_N.name: "n",
    KeyboardKey.KEY_M.name: "m",
    KeyboardKey.KEY_COMMA.name: ",",
    KeyboardKey.KEY_PERIOD.name: ".",
    KeyboardKey.KEY_SLASH.name: "/",
    KeyboardKey.KEY_RIGHT_SHIFT.name: "shiftright",
    KeyboardKey.KEY_UP.name: "up",
    KeyboardKey.KEY_NUMPAD_1.name: "num1",
    KeyboardKey.KEY_NUMPAD_2.name: "num2",
    KeyboardKey.KEY_NUMPAD_3.name: "num3",
    KeyboardKey.KEY_NUMPAD_RETURN.name: "separator",
    KeyboardKey.KEY_NUMPAD_ENTER.name: "separator",
    KeyboardKey.KEY_LEFT_CTRL.name: "ctrlleft",
    KeyboardKey.KEY_LEFT_SUPER.name: "winleft",
    KeyboardKey.KEY_LEFT_WINDOWS.name: "winleft",
    KeyboardKey.KEY_LEFT_ALT.name: "altleft",
    KeyboardKey.KEY_SPACE.name: "space",
    KeyboardKey.KEY_RIGHT_ALT.name: "altright",
    KeyboardKey.KEY_RIGHT_SUPER.name: "winright",
    KeyboardKey.KEY_RIGHT_WINDOWS.name: "winright",
    KeyboardKey.KEY_APP_MENU.name: "apps",
    KeyboardKey.KEY_RIGHT_CTRL.name: "ctrlright",
    KeyboardKey.KEY_LEFT.name: "left",
    KeyboardKey.KEY_DOWN.name: "down",
    KeyboardKey.KEY_RIGHT.name: "right",
    KeyboardKey.KEY_NUMPAD_0.name: "num0",
    KeyboardKey.KEY_NUMPAD_DECIMAL.name: "decimal",
    KeyboardKey.KEY_NUMPAD_PERIOD.name: "decimal"
}

mouse_button_mapping = {
    MouseButton.LEFT.name: "left",
    MouseButton.MIDDLE.name: "middle",
    MouseButton.RIGHT.name: "right",
}


class PyAutoGUIInputController(InputController):

    def __init__(self, game=None, **kwargs):
        self.game = game

        self.previous_key_collection_set = set()

        self.sprite_locator = SpriteLocator()

    # Keyboard Actions
    def handle_keys(self, key_collection, **kwargs):
        key_collection_set = set(key_collection)

        keys_to_press = key_collection_set - self.previous_key_collection_set
        keys_to_release = self.previous_key_collection_set - key_collection_set

        for key in keys_to_press:
            self.press_key(key)

        for key in keys_to_release:
            self.release_key(key)

        self.previous_key_collection_set = key_collection_set

    def tap_keys(self, keys, duration=0.05, **kwargs):
        if self.game_is_focused:
            for key in keys:
                self.press_key(key)

            time.sleep(duration)

            for key in keys:
                self.release_key(key)

    def tap_key(self, key, duration=0.05, **kwargs):
        if self.game_is_focused:
            self.press_key(key)

            time.sleep(duration)

            self.release_key(key)

    def press_keys(self, keys, **kwargs):
        for key in keys:
            self.press_key(key)

    def press_key(self, key, **kwargs):
        if self.game_is_focused:
            pyautogui.keyDown(keyboard_key_mapping[key.name])

    def release_keys(self, keys, **kwargs):
        for key in keys:
            self.release_key(key)

    def release_key(self, key, **kwargs):
        if self.game_is_focused:
            pyautogui.keyUp(keyboard_key_mapping[key.name])

    def type_string(self, string, duration=0.05, **kwargs):
        if self.game_is_focused:
            pyautogui.typewrite(message=string, interval=duration)

    # Mouse Actions
    def move(self, x=None, y=None, duration=0.25, absolute=True, **kwargs):
        if self.game_is_focused:
            x += self.game.window_geometry["x_offset"]
            y += self.game.window_geometry["y_offset"]

            if absolute:
                pyautogui.moveTo(x=x, y=y, duration=duration)
            else:
                pyautogui.moveRel(xOffset=x, yOffset=y, duration=duration)

    def click_down(self, button=MouseButton.LEFT, **kwargs):
        if self.game_is_focused:
            pyautogui.mouseDown(button=mouse_button_mapping[button.name])

    def click_up(self, button=MouseButton.LEFT, **kwargs):
        if self.game_is_focused:
            pyautogui.mouseUp(button=mouse_button_mapping[button.name])

    def click(self, button=MouseButton.LEFT, duration=0.25, **kwargs):
        if self.game_is_focused:
            self.click_down(button=button)
            time.sleep(duration)
            self.click_up(button=button)

    def click_screen_region(self, button=MouseButton.LEFT, screen_region=None, **kwargs):
        if self.game_is_focused:
            screen_region_coordinates = self.game.screen_regions.get(screen_region)

            x = (screen_region_coordinates[1] + screen_region_coordinates[3]) // 2
            y = (screen_region_coordinates[0] + screen_region_coordinates[2]) // 2

            self.move(x=x, y=y)
            self.click(button=button)

    def click_sprite(self, button=MouseButton.LEFT, sprite=None, game_frame=None, **kwargs):
        if self.game_is_focused:
            sprite_location = self.sprite_locator.locate(sprite=sprite, game_frame=game_frame)

            if sprite_location is None:
                return False

            x = (sprite_location[1] + sprite_location[3]) // 2
            y = (sprite_location[0] + sprite_location[2]) // 2

            self.move(x=x, y=y)
            self.click(button=button)

            return True

    def click_string(self, query_string, button=MouseButton.LEFT, game_frame=None, fuzziness=2, ocr_preset=None, **kwargs):
        if self.game_is_focused:
            string_location = serpent.ocr.locate_string(
                query_string,
                game_frame.frame,
                fuzziness=fuzziness,
                ocr_preset=ocr_preset,
                offset_x=game_frame.offset_x,
                offset_y=game_frame.offset_y
            )

            if string_location is not None:
                x = (string_location[1] + string_location[3]) // 2
                y = (string_location[0] + string_location[2]) // 2

                self.move(x=x, y=y)
                self.click(button=button)

                return True

            return False

    def drag(self, button=MouseButton.LEFT, x0=None, y0=None, x1=None, y1=None, duration=0.25, **kwargs):
        if self.game_is_focused:
            self.move(x=x0, y=y0)
            self.click_down(button=button)
            self.move(x=x1, y=y1, duration=duration)
            self.click_up(button=button)

    def drag_screen_region_to_screen_region(self, button=MouseButton.LEFT, start_screen_region=None, end_screen_region=None, duration=1, **kwargs):
        if self.game_is_focused:
            start_screen_region_coordinates = self._extract_screen_region_coordinates(start_screen_region)
            end_screen_region_coordinates = self._extract_screen_region_coordinates(end_screen_region)

            self.drag(
                button=button,
                x0=start_screen_region_coordinates[0],
                y0=start_screen_region_coordinates[1],
                x1=end_screen_region_coordinates[0],
                y1=end_screen_region_coordinates[1],
                duration=duration
            )

    def scroll(self, clicks=1, direction="DOWN", **kwargs):
        if self.game_is_focused:
            clicks = clicks * (1 if direction == "DOWN" else -1)
            pyautogui.scroll(clicks)



from kivy.app import App

from kivy.core.window import Window
from kivy.core.image import Image as CoreImage

from kivy.uix.widget import Widget
from kivy.uix.image import Image
from kivy.uix.label import Label

from kivy.uix.floatlayout import FloatLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.boxlayout import BoxLayout

from kivy.clock import Clock

from PIL import Image as PILImage

from serpent.visual_debugger.visual_debugger import VisualDebugger

from serpent.config import config

import io


class VisualDebuggerApp(App):
    def __init__(self, buckets=None):
        super().__init__()

        self.visual_debugger = VisualDebugger(buckets=buckets)
        self.canvas = None

    def build(self):
        self.canvas = VisualDebuggerCanvas(buckets=self.visual_debugger.available_buckets)

        Clock.schedule_interval(self.update_image_data, 0.01)

        return self.canvas

    def update_image_data(self, *args):
        response = self.visual_debugger.retrieve_image_data()

        if response is not None:
            bucket, image_data = response
            self.canvas.update(bucket, image_data)


class VisualDebuggerCanvas(Widget):

    def __init__(self, buckets=None):
        super().__init__()

        self.images = dict()

        self.root = FloatLayout(size=(Window.width, Window.height))
        self.grid = GridLayout(cols=8)

        self.add_widget(self.root)
        self.root.add_widget(self.grid)

        for i, bucket in enumerate(buckets):
            layout = BoxLayout(orientation="vertical")

            image = VisualDebuggerImage(
                allow_stretch=True
            )

            image.bind(texture=image.update_texture_filters)

            self.images[bucket] = image

            layout.add_widget(image)

            label = Label(
                text=bucket,
                color=(1, 1, 1, 1),
                size_hint=(1, 0.1)
            )

            layout.add_widget(label)

            self.grid.add_widget(layout)

        Window.bind(on_resize=self.on_window_resize)
        Window.clearcolor = (0.136, 0.191, 0.25, 1)

    def update(self, bucket, image_data):
        image = PILImage.fromarray(image_data).convert("RGB")
        image_file = io.BytesIO()

        image.save(image_file, "png")
        image_file.seek(0)

        core_image = CoreImage(image_file, ext="png")
        self.images[bucket].texture = core_image.texture

    def on_window_resize(self, window, width, height):
        self.root.size = (width, height)


class VisualDebuggerImage(Image):

    def update_texture_filters(self, image, texture):
        if not texture:
            return

        texture.min_filter = 'nearest'
        texture.mag_filter = 'nearest'







from serpent.input_controller import InputController, MouseButton, KeyboardKey, character_keyboard_key_mapping

from serpent.sprite_locator import SpriteLocator

import serpent.ocr

import time

import ctypes
import win32api

import scipy.interpolate
import numpy as np


keyboard_key_mapping = {
    KeyboardKey.KEY_ESCAPE.name: 0x01,
    KeyboardKey.KEY_F1.name: 0x3B,
    KeyboardKey.KEY_F2.name: 0x3C,
    KeyboardKey.KEY_F3.name: 0x3D,
    KeyboardKey.KEY_F4.name: 0x3E,
    KeyboardKey.KEY_F5.name: 0x3F,
    KeyboardKey.KEY_F6.name: 0x40,
    KeyboardKey.KEY_F7.name: 0x41,
    KeyboardKey.KEY_F8.name: 0x42,
    KeyboardKey.KEY_F9.name: 0x43,
    KeyboardKey.KEY_F10.name: 0x44,
    KeyboardKey.KEY_F11.name: 0x57,
    KeyboardKey.KEY_F12.name: 0x58,
    KeyboardKey.KEY_PRINT_SCREEN.name: 0xB7,
    KeyboardKey.KEY_SCROLL_LOCK.name: 0x46,
    KeyboardKey.KEY_PAUSE.name: 0xC5,
    KeyboardKey.KEY_GRAVE.name: 0x29,
    KeyboardKey.KEY_BACKTICK.name: 0x29,
    KeyboardKey.KEY_1.name: 0x02,
    KeyboardKey.KEY_2.name: 0x03,
    KeyboardKey.KEY_3.name: 0x04,
    KeyboardKey.KEY_4.name: 0x05,
    KeyboardKey.KEY_5.name: 0x06,
    KeyboardKey.KEY_6.name: 0x07,
    KeyboardKey.KEY_7.name: 0x08,
    KeyboardKey.KEY_8.name: 0x09,
    KeyboardKey.KEY_9.name: 0x0A,
    KeyboardKey.KEY_0.name: 0x0B,
    KeyboardKey.KEY_MINUS.name: 0x0C,
    KeyboardKey.KEY_DASH.name: 0x0C,
    KeyboardKey.KEY_EQUALS.name: 0x0D,
    KeyboardKey.KEY_BACKSPACE.name: 0x0E,
    KeyboardKey.KEY_INSERT.name: 0xD2,
    KeyboardKey.KEY_HOME.name: 0xC7,
    KeyboardKey.KEY_PAGE_UP.name: 0xC9,
    KeyboardKey.KEY_NUMLOCK.name: 0x45,
    KeyboardKey.KEY_NUMPAD_DIVIDE.name: 0xB5,
    KeyboardKey.KEY_NUMPAD_SLASH.name: 0xB5,
    KeyboardKey.KEY_NUMPAD_MULTIPLY.name: 0x37,
    KeyboardKey.KEY_NUMPAD_STAR.name: 0x37,
    KeyboardKey.KEY_NUMPAD_SUBTRACT.name: 0x4A,
    KeyboardKey.KEY_NUMPAD_DASH.name: 0x4A,
    KeyboardKey.KEY_TAB.name: 0x0F,
    KeyboardKey.KEY_Q.name: 0x10,
    KeyboardKey.KEY_W.name: 0x11,
    KeyboardKey.KEY_E.name: 0x12,
    KeyboardKey.KEY_R.name: 0x13,
    KeyboardKey.KEY_T.name: 0x14,
    KeyboardKey.KEY_Y.name: 0x15,
    KeyboardKey.KEY_U.name: 0x16,
    KeyboardKey.KEY_I.name: 0x17,
    KeyboardKey.KEY_O.name: 0x18,
    KeyboardKey.KEY_P.name: 0x19,
    KeyboardKey.KEY_LEFT_BRACKET.name: 0x1A,
    KeyboardKey.KEY_RIGHT_BRACKET.name: 0x1B,
    KeyboardKey.KEY_BACKSLASH.name: 0x2B,
    KeyboardKey.KEY_DELETE.name: 0xD3,
    KeyboardKey.KEY_END.name: 0xCF,
    KeyboardKey.KEY_PAGE_DOWN.name: 0xD1,
    KeyboardKey.KEY_NUMPAD_7.name: 0x47,
    KeyboardKey.KEY_NUMPAD_8.name: 0x48,
    KeyboardKey.KEY_NUMPAD_9.name: 0x49,
    KeyboardKey.KEY_NUMPAD_ADD.name: 0x4E,
    KeyboardKey.KEY_NUMPAD_PLUS.name: 0x4E,
    KeyboardKey.KEY_CAPSLOCK.name: 0x3A,
    KeyboardKey.KEY_A.name: 0x1E,
    KeyboardKey.KEY_S.name: 0x1F,
    KeyboardKey.KEY_D.name: 0x20,
    KeyboardKey.KEY_F.name: 0x21,
    KeyboardKey.KEY_G.name: 0x22,
    KeyboardKey.KEY_H.name: 0x23,
    KeyboardKey.KEY_J.name: 0x24,
    KeyboardKey.KEY_K.name: 0x25,
    KeyboardKey.KEY_L.name: 0x26,
    KeyboardKey.KEY_SEMICOLON.name: 0x27,
    KeyboardKey.KEY_APOSTROPHE.name: 0x28,
    KeyboardKey.KEY_RETURN.name: 0x1C,
    KeyboardKey.KEY_ENTER.name: 0x1C,
    KeyboardKey.KEY_NUMPAD_4.name: 0x4B,
    KeyboardKey.KEY_NUMPAD_5.name: 0x4C,
    KeyboardKey.KEY_NUMPAD_6.name: 0x4D,
    KeyboardKey.KEY_LEFT_SHIFT.name: 0x2A,
    KeyboardKey.KEY_Z.name: 0x2C,
    KeyboardKey.KEY_X.name: 0x2D,
    KeyboardKey.KEY_C.name: 0x2E,
    KeyboardKey.KEY_V.name: 0x2F,
    KeyboardKey.KEY_B.name: 0x30,
    KeyboardKey.KEY_N.name: 0x31,
    KeyboardKey.KEY_M.name: 0x32,
    KeyboardKey.KEY_COMMA.name: 0x33,
    KeyboardKey.KEY_PERIOD.name: 0x34,
    KeyboardKey.KEY_SLASH.name: 0x35,
    KeyboardKey.KEY_RIGHT_SHIFT.name: 0x36,
    KeyboardKey.KEY_UP.name: 0xC8,
    KeyboardKey.KEY_NUMPAD_1.name: 0x4F,
    KeyboardKey.KEY_NUMPAD_2.name: 0x50,
    KeyboardKey.KEY_NUMPAD_3.name: 0x51,
    KeyboardKey.KEY_NUMPAD_RETURN.name: 0x9C,
    KeyboardKey.KEY_NUMPAD_ENTER.name: 0x9C,
    KeyboardKey.KEY_LEFT_CTRL.name: 0x1D,
    KeyboardKey.KEY_LEFT_SUPER.name: 0xDB,
    KeyboardKey.KEY_LEFT_WINDOWS.name: 0xDB,
    KeyboardKey.KEY_LEFT_ALT.name: 0x38,
    KeyboardKey.KEY_SPACE.name: 0x39,
    KeyboardKey.KEY_RIGHT_ALT.name: 0xB8,
    KeyboardKey.KEY_RIGHT_SUPER.name: 0xDC,
    KeyboardKey.KEY_RIGHT_WINDOWS.name: 0xDC,
    KeyboardKey.KEY_APP_MENU.name: 0xDD,
    KeyboardKey.KEY_RIGHT_CTRL.name: 0x9D,
    KeyboardKey.KEY_LEFT.name: 0xCB,
    KeyboardKey.KEY_DOWN.name: 0xD0,
    KeyboardKey.KEY_RIGHT.name: 0xCD,
    KeyboardKey.KEY_NUMPAD_0.name: 0x52,
    KeyboardKey.KEY_NUMPAD_DECIMAL.name: 0x53,
    KeyboardKey.KEY_NUMPAD_PERIOD.name: 0x53
}

mouse_button_down_mapping = {
    MouseButton.LEFT.name: 0x0002,
    MouseButton.MIDDLE.name: 0x0020,
    MouseButton.RIGHT.name: 0x0008
}

mouse_button_up_mapping = {
    MouseButton.LEFT.name: 0x0004,
    MouseButton.MIDDLE.name: 0x0040,
    MouseButton.RIGHT.name: 0x0010
}

PUL = ctypes.POINTER(ctypes.c_ulong)


class KeyBdInput(ctypes.Structure):
    _fields_ = [("wVk", ctypes.c_ushort),
                ("wScan", ctypes.c_ushort),
                ("dwFlags", ctypes.c_ulong),
                ("time", ctypes.c_ulong),
                ("dwExtraInfo", PUL)]


class HardwareInput(ctypes.Structure):
    _fields_ = [("uMsg", ctypes.c_ulong),
                ("wParamL", ctypes.c_short),
                ("wParamH", ctypes.c_ushort)]


class MouseInput(ctypes.Structure):
    _fields_ = [("dx", ctypes.c_long),
                ("dy", ctypes.c_long),
                ("mouseData", ctypes.c_ulong),
                ("dwFlags", ctypes.c_ulong),
                ("time", ctypes.c_ulong),
                ("dwExtraInfo", PUL)]


class Input_I(ctypes.Union):
    _fields_ = [("ki", KeyBdInput),
                ("mi", MouseInput),
                ("hi", HardwareInput)]


class Input(ctypes.Structure):
    _fields_ = [("type", ctypes.c_ulong),
                ("ii", Input_I)]


class NativeWin32InputController(InputController):

    def __init__(self, game=None, **kwargs):
        self.game = game

        self.previous_key_collection_set = set()

        self.sprite_locator = SpriteLocator()

    # Keyboard Actions
    def handle_keys(self, key_collection, **kwargs):
        key_collection_set = set(key_collection)

        keys_to_press = key_collection_set - self.previous_key_collection_set
        keys_to_release = self.previous_key_collection_set - key_collection_set

        for key in keys_to_press:
            self.press_key(key)

        for key in keys_to_release:
            self.release_key(key)

        self.previous_key_collection_set = key_collection_set

    def tap_keys(self, keys, duration=0.05, **kwargs):
        if self.game_is_focused:
            for key in keys:
                self.press_key(key)

            time.sleep(duration)

            for key in keys:
                self.release_key(key)

    def tap_key(self, key, duration=0.05, **kwargs):
        if self.game_is_focused:
            self.press_key(key)

            time.sleep(duration)

            self.release_key(key)

    def press_keys(self, keys, **kwargs):
        for key in keys:
            self.press_key(key)

    def press_key(self, key, **kwargs):
        if self.game_is_focused:
            extra = ctypes.c_ulong(0)
            ii_ = Input_I()
            ii_.ki = KeyBdInput(0, keyboard_key_mapping[key.name], 0x0008, 0, ctypes.pointer(extra))
            x = Input(ctypes.c_ulong(1), ii_)
            ctypes.windll.user32.SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    def release_keys(self, keys, **kwargs):
        for key in keys:
            self.release_key(key)

    def release_key(self, key, **kwargs):
        if self.game_is_focused:
            extra = ctypes.c_ulong(0)
            ii_ = Input_I()
            ii_.ki = KeyBdInput(0, keyboard_key_mapping[key.name], 0x0008 | 0x0002, 0, ctypes.pointer(extra))
            x = Input(ctypes.c_ulong(1), ii_)
            ctypes.windll.user32.SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    def type_string(self, string, duration=0.05, **kwargs):
        if self.game_is_focused:
            for character in string:
                keys = character_keyboard_key_mapping.get(character)

                if keys is not None:
                    self.tap_keys(keys, duration=duration)

    # Mouse Actions
    def move(self, x=None, y=None, duration=0.25, absolute=True, **kwargs):
        if self.game_is_focused:
            x += self.game.window_geometry["x_offset"]
            y += self.game.window_geometry["y_offset"]

            current_pixel_coordinates = win32api.GetCursorPos()
            start_coordinates = self._to_windows_coordinates(*current_pixel_coordinates)

            if absolute:
                end_coordinates = self._to_windows_coordinates(x, y)
            else:
                end_coordinates = self._to_windows_coordinates(current_pixel_coordinates[0] + x, current_pixel_coordinates[1] + y)

            coordinates = self._interpolate_mouse_movement(
                start_windows_coordinates=start_coordinates,
                end_windows_coordinates=end_coordinates
            )

            for x, y in coordinates:
                extra = ctypes.c_ulong(0)
                ii_ = Input_I()
                ii_.mi = MouseInput(x, y, 0, (0x0001 | 0x8000), 0, ctypes.pointer(extra))
                x = Input(ctypes.c_ulong(0), ii_)
                ctypes.windll.user32.SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

                time.sleep(duration / 20)

    def click_down(self, button=MouseButton.LEFT, **kwargs):
        if self.game_is_focused:
            extra = ctypes.c_ulong(0)
            ii_ = Input_I()
            ii_.mi = MouseInput(0, 0, 0, mouse_button_down_mapping[button.name], 0, ctypes.pointer(extra))
            x = Input(ctypes.c_ulong(0), ii_)
            ctypes.windll.user32.SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    def click_up(self, button=MouseButton.LEFT, **kwargs):
        if self.game_is_focused:
            extra = ctypes.c_ulong(0)
            ii_ = Input_I()
            ii_.mi = MouseInput(0, 0, 0, mouse_button_up_mapping[button.name], 0, ctypes.pointer(extra))
            x = Input(ctypes.c_ulong(0), ii_)
            ctypes.windll.user32.SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    def click(self, button=MouseButton.LEFT, duration=0.05, **kwargs):
        if self.game_is_focused:
            self.click_down(button=button)
            time.sleep(duration)
            self.click_up(button=button)

    def click_screen_region(self, button=MouseButton.LEFT, screen_region=None, **kwargs):
        if self.game_is_focused:
            screen_region_coordinates = self.game.screen_regions.get(screen_region)

            x = (screen_region_coordinates[1] + screen_region_coordinates[3]) // 2
            y = (screen_region_coordinates[0] + screen_region_coordinates[2]) // 2

            self.move(x, y)
            self.click(button=button)

    def click_sprite(self, button=MouseButton.LEFT, sprite=None, game_frame=None, **kwargs):
        if self.game_is_focused:
            sprite_location = self.sprite_locator.locate(sprite=sprite, game_frame=game_frame)

            if sprite_location is None:
                return False

            x = (sprite_location[1] + sprite_location[3]) // 2
            y = (sprite_location[0] + sprite_location[2]) // 2

            self.move(x, y)
            self.click(button=button)

            return True

    def click_string(self, query_string, button=MouseButton.LEFT, game_frame=None, fuzziness=2, ocr_preset=None, **kwargs):
        if self.game_is_focused:
            string_location = serpent.ocr.locate_string(
                query_string,
                game_frame.frame,
                fuzziness=fuzziness,
                ocr_preset=ocr_preset,
                offset_x=game_frame.offset_x,
                offset_y=game_frame.offset_y
            )

            if string_location is not None:
                x = (string_location[1] + string_location[3]) // 2
                y = (string_location[0] + string_location[2]) // 2

                self.move(x, y)
                self.click(button=button)

                return True

            return False

    def drag(self, button=MouseButton.LEFT, x0=None, y0=None, x1=None, y1=None, duration=0.25, **kwargs):
        if self.game_is_focused:
            self.move(x=x0, y=y0)
            self.click_down(button=button)
            self.move(x=x1, y=y1, duration=duration)
            self.click_up(button=button)

    def drag_screen_region_to_screen_region(self, button=MouseButton.LEFT, start_screen_region=None, end_screen_region=None, duration=1, **kwargs):
        if self.game_is_focused:
            start_screen_region_coordinates = self._extract_screen_region_coordinates(start_screen_region)
            end_screen_region_coordinates = self._extract_screen_region_coordinates(end_screen_region)

            self.drag(
                button=button,
                x0=start_screen_region_coordinates[0],
                y0=start_screen_region_coordinates[1],
                x1=end_screen_region_coordinates[0],
                y1=end_screen_region_coordinates[1],
                duration=duration
            )

    def scroll(self, clicks=1, direction="DOWN", **kwargs):
        if self.game_is_focused:
            clicks = clicks * (1 if direction == "UP" else -1) * 120

            extra = ctypes.c_ulong(0)
            ii_ = Input_I()
            ii_.mi = MouseInput(0, 0, clicks, 0x0800, 0, ctypes.pointer(extra))
            x = Input(ctypes.c_ulong(0), ii_)
            ctypes.windll.user32.SendInput(1, ctypes.pointer(x), ctypes.sizeof(x))

    @staticmethod
    def _to_windows_coordinates(x=0, y=0):
        display_width = win32api.GetSystemMetrics(0)
        display_height = win32api.GetSystemMetrics(1)

        windows_x = (x * 65535) // display_width
        windows_y = (y * 65535) // display_height

        return windows_x, windows_y

    @staticmethod
    def _interpolate_mouse_movement(start_windows_coordinates, end_windows_coordinates, steps=20):
        x_coordinates = [start_windows_coordinates[0], end_windows_coordinates[0]]
        y_coordinates = [start_windows_coordinates[1], end_windows_coordinates[1]]

        if x_coordinates[0] == x_coordinates[1]:
            x_coordinates[1] += 1

        if y_coordinates[0] == y_coordinates[1]:
            y_coordinates[1] += 1

        interpolation_func = scipy.interpolate.interp1d(x_coordinates, y_coordinates)

        intermediate_x_coordinates = np.linspace(start_windows_coordinates[0], end_windows_coordinates[0], steps + 1)[1:]
        coordinates = list(map(lambda x: (int(round(x)), int(interpolation_func(x))), intermediate_x_coordinates))

        return coordinates



from redis import StrictRedis

import skimage.io
import numpy as np

import pickle
import itertools

from serpent.config import config


class VisualDebugger:

    def __init__(self, buckets=None):
        self.available_buckets = buckets or config["visual_debugger"]["available_buckets"]
        self.bucket_generator = itertools.cycle(self.available_buckets)

        self.redis_client = StrictRedis(**config["redis"])
        self.clear_image_data()

    def store_image_data(self, image_data, image_shape, bucket="debug"):
        self.redis_client.lpush(f"{config['visual_debugger']['redis_key_prefix']}:{bucket}:SHAPE", pickle.dumps(image_shape))
        self.redis_client.lpush(f"{config['visual_debugger']['redis_key_prefix']}:{bucket}", image_data.tobytes())

    def retrieve_image_data(self):
        bucket = next(self.bucket_generator)
        bucket_key = f"{config['visual_debugger']['redis_key_prefix']}:{bucket}"

        response = self.redis_client.rpop(bucket_key)

        if response is not None:
            bucket = bucket_key.split(":")[-1]

            image_shape = self.redis_client.rpop(f"{config['visual_debugger']['redis_key_prefix']}:{bucket}:SHAPE")
            image_shape = pickle.loads(image_shape)

            image_data = np.fromstring(response, dtype="uint8").reshape(image_shape)

            return bucket, image_data

        return None

    def save_image_data(self, bucket, image_data):
        if bucket in self.available_buckets:
            if image_data.dtype == "bool" or (image_data.dtype == "uint8" and 1 in np.unique(image_data)):
                image_data = image_data.astype("uint8") * 255

            skimage.io.imsave(f"{bucket}.png", image_data)

    def clear_image_data(self):
        visual_debugger_keys = self.redis_client.keys(f"{config['visual_debugger']['redis_key_prefix']}*")

        for key in visual_debugger_keys:
            self.redis_client.delete(key.decode("utf-8"))

    def get_bucket_queue_length(self, bucket):
        return self.redis_client.llen(f"{config['visual_debugger']['redis_key_prefix']}:{bucket}")



import serpent.datasets

import os
import random

import skimage.io


class ContextClassifierError(BaseException):
    pass


class ContextClassifier:
    def __init__(self, **kwargs):
        self.classifier = None

    @property
    def training_sample_count(self):
        sample_count = 0

        for root, dirs, files in os.walk("datasets/current/training"):
            sample_count += len(files)

        return sample_count

    @property
    def validation_sample_count(self):
        sample_count = 0

        for root, dirs, files in os.walk("datasets/current/validation"):
            sample_count += len(files)

        return sample_count

    @staticmethod
    def create_training_and_validation_sets(validation_set_probability=0.1, seed=None):
        class_directories = None

        for root, dirs, files in os.walk("datasets/collect_frames_for_context"):
            class_directories = [f"{root}/{d}" for d in dirs]
            break

        serpent.datasets.create_training_and_validation_sets(
            class_directories,
            validation_set_probability=validation_set_probability,
            seed=seed
        )

    def train(self, epochs=3):
        raise NotImplementedError()

    def validate(self):
        raise NotImplementedError()

    def predict(self, input_frame):
        raise NotImplementedError()

    def save_classifier(self, file_path):
        raise NotImplementedError()

    def load_classifier(self, file_path):
        raise NotImplementedError()

    @classmethod
    def available_implementations(cls):
        return [
            "SVMContextClassifier",
            "CNNInceptionV3ContextClassifier"
        ]

    @classmethod
    def executable_train(cls, epochs=3, autosave=False, classifier="CNNInceptionV3ContextClassifier", validate=True):
        context_paths = list()

        for root, directories, files in os.walk("datasets/collect_frames_for_context".replace("/", os.sep)):
            if root != "datasets/collect_frames_for_context".replace("/", os.sep):
                break

            for directory in directories:
                context_paths.append(f"datasets/collect_frames_for_context/{directory}".replace("/", os.sep))

        if not len(context_paths):
            raise ContextClassifierError("No Context Frames found in 'datasets/collect_frames_for_datasets'...")

        serpent.datasets.create_training_and_validation_sets(context_paths)

        context_classifier_class = cls.context_classifier_mapping().get(classifier)

        if context_classifier_class is not None:
            context_path = random.choice(context_paths)
            frame_path = None

            for root, directories, files in os.walk(context_path):
                for file in files:
                    if file.endswith(".png"):
                        frame_path = f"{context_path}/{file}"
                        break

                if frame_path is not None:
                    break

            frame = skimage.io.imread(frame_path)

            context_classifier = context_classifier_class(input_shape=frame.shape)

            context_classifier.train(epochs=epochs, autosave=autosave, validate=validate)
            context_classifier.validate()

            context_classifier.save_classifier("datasets/context_classifier.model")
            print("Success! Model was saved to 'datasets/context_classifier.model'")

    @classmethod
    def context_classifier_mapping(cls):
        from serpent.machine_learning.context_classification.context_classifiers.svm_context_classifier import SVMContextClassifier
        from serpent.machine_learning.context_classification.context_classifiers.cnn_inception_v3_context_classifier import CNNInceptionV3ContextClassifier

        return {
            "SVMContextClassifier": SVMContextClassifier,
            "CNNInceptionV3ContextClassifier": CNNInceptionV3ContextClassifier
        }




import numpy as np

import mss

from redis import StrictRedis

import skimage.transform
import skimage.util

from serpent.config import config

import sys
import time

from datetime import datetime

from serpent.game_frame import GameFrame
from serpent.game_frame_buffer import GameFrameBuffer


redis_client = StrictRedis(**config["redis"])


class FrameGrabber:

    def __init__(self, width=640, height=480, x_offset=0, y_offset=0, fps=30, buffer_seconds=5):
        self.width = width
        self.height = height

        self.x_offset = x_offset
        self.y_offset = y_offset

        self.frame_time = 1 / fps
        self.frame_buffer_size = buffer_seconds * fps

        self.redis_client = redis_client
        self.screen_grabber = mss.mss()

        self.is_retina_display = False
        self.is_retina_display = self._perform_retina_display_check()

        # Clear any previously stored frames
        self.redis_client.delete(config["frame_grabber"]["redis_key"])

    def start(self):
        while True:
            cycle_start = datetime.utcnow()
            frame = self.grab_frame()

            self.redis_client.lpush(config["frame_grabber"]["redis_key"], frame.tobytes())
            self.redis_client.ltrim(config["frame_grabber"]["redis_key"], 0, self.frame_buffer_size)

            cycle_end = datetime.utcnow()
            cycle_duration = (cycle_end - cycle_start).microseconds / 1000000

            frame_time_left = self.frame_time - cycle_duration

            if frame_time_left > 0:
                time.sleep(frame_time_left)

    def grab_frame(self):
        frame = np.array(
            self.screen_grabber.grab({
                "top": self.y_offset,
                "left": self.x_offset,
                "width": self.width,
                "height": self.height
            }),
            dtype="uint8"
        )

        frame = frame[..., [2, 1, 0, 3]]

        if self.is_retina_display:
            frame = skimage.util.img_as_ubyte(skimage.transform.resize(frame, (frame.shape[0] // 2, frame.shape[1] // 2)))
            frame = frame[:self.height, :self.width, :]

        return frame[..., :3]

    def _perform_retina_display_check(self):
        retina_display = False

        if sys.platform == "darwin":
            frame = self.grab_frame()

            if frame.shape[0] > self.height:
                retina_display = True

        return retina_display


    @classmethod
    def get_frames(cls, frame_buffer_indices, frame_shape=None):
        game_frame_buffer = GameFrameBuffer(size=len(frame_buffer_indices))

        for i in frame_buffer_indices:
            frame_bytes = redis_client.lindex(config["frame_grabber"]["redis_key"], i)
            frame_array = np.fromstring(frame_bytes, dtype="uint8").reshape(frame_shape)

            game_frame = GameFrame(frame_array)

            game_frame_buffer.add_game_frame(game_frame)

        return game_frame_buffer



import numpy as np

import skimage.io
import skimage.util

import os


def extract_region_from_image(image, region_bounding_box):
    return image[region_bounding_box[0]:region_bounding_box[2], region_bounding_box[1]:region_bounding_box[3]]


def isolate_sprite(image_region_path, output_file_path):
    result_image = None

    for root, directories, files in os.walk(image_region_path):
        for file in files:
            if not file.endswith(".png"):
                continue

            image = skimage.io.imread(f"{root}/{file}")
            image = np.concatenate((image, np.full((image.shape[0], image.shape[1], 1), 255, dtype="uint8")), axis=2)

            if result_image is None:
                result_image = image
            else:
                height, width, rgba = image.shape

                for i in range(height):
                    for ii in range(width):
                        if not np.array_equal(image[i, ii, :2], result_image[i, ii, :2]):
                            result_image[i, ii, 3] = 0

    skimage.io.imsave(output_file_path, result_image)


def scale_range(n, minimum, maximum):
    n += -(np.min(n))
    n /= np.max(n) / (maximum - minimum)
    n += minimum

    return n



from serpent.game_launcher import GameLauncher, GameLauncherException

import sys
import shlex
import subprocess


class ExecutableGameLauncher(GameLauncher):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def launch(self, **kwargs):
        executable_path = kwargs.get("executable_path")

        if executable_path is None:
            raise GameLauncherException("An 'executable_path' kwarg is required...")

        if sys.platform in ["linux", "linux2"]:
            subprocess.Popen(shlex.split(executable_path))
        elif sys.platform == "darwin":
            subprocess.Popen(shlex.split(executable_path))
        elif sys.platform == "win32":
            subprocess.Popen(shlex.split(executable_path))